/*
 * concatenation.cpp
 *
 *  Created on: Aug 11, 2017
 *      Author: julien
 */

#include "concatenation.hpp"

Concatenation::Concatenation(unsigned short gradient_threshold,
		  	  	  	  	     int line_distance_threshold,
							 int line_spread,
							 int min_line_length,
							 int translation_x,
							 int translation_y):
	translation(translation_x, translation_y),
	gradient_threshold(gradient_threshold),
	line_distance_threshold(line_distance_threshold),
	line_spread(line_spread*line_spread),
	min_line_length(min_line_length)
{
}

void Concatenation::concatenate(Mat* grad_x,
								Mat* grad_y,
								vector<Point>* contour_l,
								vector<Point>* contour_r){



	lookup_x = *grad_x;
	lookup_y = *grad_y;

	collect_points();

	concat();

//	cout << "check" << endl;

//	temp_contours.clear();
//	temp_contours.push_back(structured_points_l);
//	temp_contours.push_back(structured_points_r);

	filter();

//	cout << "check2" << endl;

//	*contours = structured_points_l;
//	*contours = right_contours;
//	*contours = temp_contours.front();

	*contour_l = left_contour;
	*contour_r = right_contour;

//	cout << left_contour << endl;
//	cout << right_contour << endl;

}

void Concatenation::concat(){

	temp_contours.clear();

//	thread concat_left(&Concatenation::concat_helper,
//					   this ,
//					   &structured_points_l,
//					   &left_contours,
//					   left);
//
//	thread concat_right(&Concatenation::concat_helper,
//					    this,
//						&structured_points_r,
//						&right_contours,
//						right);
//
//	concat_left.join();
//	concat_right.join();


	concat_helper(&structured_points_l,
				  &left_contour,
				  left);

	concat_helper(&structured_points_r,
				  &right_contour,
				  right);


}



void Concatenation::merge(){


}


struct less_than
{
    inline bool operator() (const Point& p1, const Point& p2)
    {
        return (p1.y > p2.y);
    }
};


void Concatenation::filter(){

	filtercontour(&left_contour, 20);

	filtercontour(&right_contour, 20);


	// sorting for the right side is needed, since the point are collected a little
	// differently
	std::sort(right_contour.begin(), right_contour.end(), less_than());

}

// TODO member funktion daraus machen
void Concatenation::add_nearest(vector<Point>* contour, Point reference_point, vector<Point>* lane_contour){
	// the first nearest point is at index 0
	Point nearest_point = contour->at(0);
	Point gradient_dif;
	int sub_vector_size = contour->size();
	unsigned int current_distance;
	unsigned int nearest_distance;

	//for all Points inside the sub vector
	for(int i = 1; i < sub_vector_size; i++){
		// if the absolute value of the current nearest point.x - reference is smaller than the
		// absolute value of the difference between the currently looked at point and the reference,
		// then nearest point remains the same, otherwise overwrite the nearest Point with the other Point

		current_distance = radial_distance(contour->at(i), reference_point);

		if(current_distance < nearest_distance){
			nearest_point = contour->at(i);
			nearest_distance = current_distance;
		}

	}

	lane_contour->push_back(nearest_point);
}


void Concatenation::concat_helper(vector<vector<Point>>* structured_points,
								  vector<Point>* lane_contour,
								  side s){

	lane_contour->clear();

	int structured_size = structured_points->size();

	// we don't need to do anything, if the points are not even there to concatenate
	if(structured_size == 0){
		return;
	}

	// first sift out all the points that are too close to each other (on the x-axis)
	sift_out(structured_points, s);

	// we start looking into the first sub vector since the overlying structure is
	// sorted descending by y-value (after sift_out)
	// and we need to start at the bottom of the picture
	Point start; //TODO in klassenvariable umwandeln und automatisch vor generieren
	// seed the starting point
	if(s == left){
		start = Point(44, 480);
	}
	else{
		start = Point(700, 480);
	}

	add_nearest(&structured_points->at(0), start, lane_contour);

	for(int i = 1; i < structured_size; i++){
		add_nearest(&structured_points->at(i), lane_contour->back(), lane_contour);
	}
}



void Concatenation::continue_lines(vector<vector<Point>>* lines, vector<Point>* next_points){

	vector<vector<Point>> temp_lines;

	int best_point; // the index of the currently best fitting point
	double best_distance;
	vector<int> best_points; // the indexes of the best points

	 // index to the currently looked at point
	double distance; // the distance to the current point


	int lines_size = lines->size();
	int points_size = next_points->size();

	// points to the last point of a line
	Point* line_last;

	// go through all the lines
	for(int l = 0; l < lines_size; l++){

		line_last = &(lines->at(l).back());

		best_distance = -1;

		for(int point = 0; point < points_size; point++){

			// first check if we are in an acceptable range (x-wise)
			if(abs(next_points->at(point).x - line_last->x) < line_spread){

				// if there has been no best distance yet,
				if(best_distance == -1){
					// we assume the current point is the best choice
					best_distance = radial_distance(next_points->at(point), *line_last);
					best_point = point;
					continue;
				}

				distance = radial_distance(next_points->at(point), *line_last);

				// if the point is closer
				if(distance < best_distance){
					// we make that the best choice
					best_distance = distance;
					best_point = point;
				}
			}
		}

		// if we got a new point to add to our current line
		if(best_distance != -1){
			// add the best point to the line
			lines->at(l).push_back(next_points->at(best_point));
		}

		// also note down the index of the best point
		best_points.push_back(best_point);
	}

	// think of the following code as a kind of riffle shuffle


	Point* point_ptr;
	int l = 0;
	// now look at all the indexes we didn't note down and start a new line with these points
	for(int point = 0; point < points_size; point++){
		// skip all those, we already found a pair for
		if(find(best_points.begin(), best_points.end(), point) != best_points.end()){
			continue;
		}
		// they have to be put in the right order (with the lines going from left to right)
		// therefore we need to find the right index to insert into

		point_ptr = &next_points->at(point);

		// we always continue on the last line we looked at
		for(; l < lines_size; l++){
			// if the x-value of the Point is smaller then the x-value of the last Point of the line
			if(point_ptr->x < lines->at(l).back().x){
				// create a new line
				temp_lines.push_back(vector<Point>());
				// and insert the Point into it (basicly creating a new starting point)
				temp_lines.back().push_back(*point_ptr);
				// we then can skip out of the loop, since the Point already has found it's place
				break;
			}
			else{
				temp_lines.push_back(lines->at(l));
			}
		}
		// if we already used all the available lines, we still need to insert the other points
		// and start a new line with each of them
		if(l >= lines_size){
			// create a new line
			temp_lines.push_back(vector<Point>());
			// and insert the Point into it (basicly creating a new starting point)
			temp_lines.back().push_back(*point_ptr);
		}
	}

	// if we used up all the points, but there are still lines left over
	// we still need to insert all the left over lines
	for(; l < lines_size; l++){
		temp_lines.push_back(lines->at(l));
	}

	// now "return" the temporary vector
	*lines = temp_lines;

}


void Concatenation::sift_out(vector<vector<Point>>* structured_points, side s){

	vector<vector<Point>> temp_structure;

	int size = structured_points->size();
	vector<Point>* row;

	// depending on what side we start looking, we delete other points
	if(s == right){
		for(int i = 0; i < size; i++){ // this loop goes through the y-values

			row = &structured_points->at(i);

			temp_structure.push_back(vector<Point>());
			temp_structure.at(i).push_back(row->front());

			for(int j = 1; j < row->size(); j++){
				if(abs(temp_structure.at(i).back().x - row->at(j).x) >= line_distance_threshold){
					temp_structure.at(i).push_back(row->at(j));
				}
			}
		}
	}
	else if(s == left){
		for(int i = 0; i < size; i++){ // this loop goes through the y-values

			row = &structured_points->at(i);

			temp_structure.push_back(vector<Point>());
			temp_structure.at(i).push_back(row->back());


			for(int j = row->size() - 2; j >= 0; j--){
				if(abs(temp_structure.at(i).back().x - row->at(j).x) >= line_distance_threshold){
					temp_structure.at(i).push_back(row->at(j));
				}
			}
		}
		// now we nee to flip the vector, since we added
		// the Point's from left and not from right.
		// this is important for other methods!
		reverse(temp_structure.begin(),temp_structure.end());
	}

	*structured_points = temp_structure;

}


void Concatenation::filtercontour(vector<Point>* contour, int lookahead){


	int iteration_range = contour->size() - lookahead;

	if(iteration_range < 1) return;


	int contour_size = contour->size();
	vector<Point> temp_contour;

	Point* reference_point;
	int closest_index; // the index of the closest calculated point

	int closest_distance;
	int current_distance;

	int reference_index = 0;	// look at the first index
	// and save that Point inside the temporary vector temp_contour
	temp_contour.push_back(contour->at(reference_index));
	while(reference_index < iteration_range){
		// keep the point in mind
		reference_point = &contour->at(reference_index);
		// make the next point in line the closest for now
		closest_index = reference_index + 1;
		// and calculate the distance between the reference point an the next
		closest_distance = radial_distance(*reference_point, contour->at(closest_index));

		// look at the next points after that (but don't go out of bounds)
		for(int j = reference_index + 2; (j < reference_index + lookahead) && (j < contour_size); j++){
			// see if this point is closer than the current closest
			current_distance = radial_distance(*reference_point, contour->at(j));
			if(current_distance < closest_distance){
				// and if so, then save the point and the distance as the closest
				closest_index = j;
				closest_distance = current_distance;
			}
		}

		// save the closest point inside the temporary vector
		temp_contour.push_back(contour->at(closest_index));
		// overwrite the reference index so we will continue from the new closest point
		reference_index = closest_index;
	}

	// overwrite the old contour with the temporary vector
	*contour = temp_contour;

}


void Concatenation::collect_points(){
	// clear out leftover Information from the last run
	structured_points_l.clear();
	structured_points_r.clear();

	short* row_pointer_x_gradient;
	short* row_pointer_y_gradient;

	int current_y_l = 0;
	int current_y_r = 0;
	int current_sub_vector_l = 0;
	int current_sub_vector_r = 0;
	Point current_point;


	// we will look at the edges-mat
	// go through all rows
	for(int row = 0; row < lookup_x.rows; row++){

		row_pointer_x_gradient = lookup_x.ptr<short>(row);
		row_pointer_y_gradient = lookup_y.ptr<short>(row);

		// go through all columns
		for(int col = 0; col < lookup_x.cols; col++){

			// if all the gradients are 0
			if(row_pointer_x_gradient[col] == 0 && row_pointer_y_gradient[col] == 0 ){
				// then we do not need to look at it
				continue;
			}

			current_point = Point(col + translation.x, row + translation.y);

			if(row_pointer_x_gradient[col] == 0){
				continue;
			}

			// if the point is on the left side of the roi
			if(current_point.x < 450){//TODO
				// and the gradient is smaller than the small threshold
				if(row_pointer_x_gradient[col] < (-1 * gradient_threshold)){
					// add it to the corresponding sorted vector
					add_to_vec(&structured_points_l, &current_point, &current_sub_vector_l, &current_y_l);
					continue;
				}
				// from here we can skip the rest of the loop

			}

			// if the point is on the right side of the roi
			else if(current_point.x > 294){//TODO
				// and the gradient is bigger than the positive threshold
				if(row_pointer_x_gradient[col] > gradient_threshold){
					// add it to the corresponding sorted vector
					add_to_vec(&structured_points_r, &current_point, &current_sub_vector_r, &current_y_r);
				}
			}
		}
	}
}

void Concatenation::add_to_vec(vector<vector<Point>>* structured_points,
							   Point* current_point,
							   int* current_sub_vector,
							   int* current_y){
	// if the next Point has a different y-value than the currently saved
	if(*current_y != current_point->y){
		// overwrite old current_y
		*current_y = current_point->y;
		// create new vector inside structured_points just for that y-value
		structured_points->push_back(vector<Point>());
		// Determine the new index of the newly created sub_vector
		*current_sub_vector = (*structured_points).size() - 1;
	}

	// put the currently looked at point into the currently looked at sub vector
	(*structured_points)[*current_sub_vector].push_back(*current_point);
}

double Concatenation::radial_distance(Point a, Point b){
	int dx = abs(a.x-b.x);
	int dy = abs(a.y-b.y);
	double distance = dx*dx + dy*dy;
	return distance;
}
